/*
  Copyright 2018-2021, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2021, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  This file is part of GekkoFS.

  GekkoFS is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GekkoFS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GekkoFS.  If not, see <https://www.gnu.org/licenses/>.

  SPDX-License-Identifier: GPL-3.0-or-later
*/

#ifndef GEKKOFS_API_DETAIL_CONNECTION_IMPL_HPP
#define GEKKOFS_API_DETAIL_CONNECTION_IMPL_HPP

#include <memory>
#include <vector>
#include <hermes.hpp>

// forward declarations
namespace gkfs::metadata {
class Metadata;
}

namespace gkfs::rpc {
class Distributor;
}

namespace gkfs::api::fs {

namespace detail {

struct peer_list {

    using endpoint_type = hermes::endpoint;

    peer_list(uint64_t localhost_id, std::vector<endpoint_type> endpoints)
        : localhost_id_(localhost_id), endpoints_(std::move(endpoints)) {}

    uint64_t
    localhost_id() const {
        return localhost_id_;
    }

    hermes::endpoint
    localhost() const {
        return endpoints_.at(localhost_id_);
    }

    std::size_t
    size() const {
        return endpoints_.size();
    }

    hermes::endpoint
    at(std::size_t n) const {
        return endpoints_.at(n);
    }

    uint64_t localhost_id_ = 0;
    std::vector<hermes::endpoint> endpoints_;
};

} // namespace detail

class connection::impl {

public:
    explicit impl(const std::string_view& hostfile);
    impl(const impl& other) = delete;
    impl(impl&& rhs) noexcept;
    impl&
    operator=(const impl& other) = delete;
    impl&
    operator=(impl&& rhs) noexcept;
    ~impl();

    void
    put(const void* buffer, size_t size,
        const std::filesystem::path& gkfs_filename) const;

    std::string
    stat(const std::string& path) const;

    void
    create(const std::string& gkfs_filename, mode_t mode) const;

    ssize_t
    pwrite(const std::string& gkfs_filename, const void* buffer, size_t size,
           off_t offset) const;

private:
    void
    check_parent_directory(const std::string& path) const;

    std::optional<gkfs::metadata::Metadata>
    fetch_metadata(const std::string& path, bool follow_links = true) const;

    int
    forward_create_rpc(const std::string& gkfs_filename,
                       const mode_t mode) const;

    std::tuple<int, off64_t>
    forward_update_metadentry_size_rpc(const std::string& gkfs_filename,
                                       size_t size, off_t offset,
                                       bool is_append) const;

    std::tuple<int, ssize_t>
    forward_write_rpc(const std::string& gkfs_filename, const void* buffer,
                      off_t offset, size_t write_size, size_t updated_metadentry_size,
                      bool is_append) const;

private:
    ////////////////////////////////////////////////////////////////////////////
    // members //
    ////////////////////////////////////////////////////////////////////////////
    std::vector<fs::host> hosts_;
    rpc::config rpc_config_;
    std::unique_ptr<hermes::async_engine> rpc_service_;
    detail::peer_list peers_;
    std::vector<hermes::endpoint> host_endpoints_;
    fs::config fs_config_;
    std::unique_ptr<gkfs::rpc::Distributor> distributor_;
};

} // namespace gkfs::api::fs

#endif // GEKKOFS_API_DETAIL_CONNECTION_IMPL_HPP