/*
  Copyright 2018-2021, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2021, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  This file is part of GekkoFS.

  GekkoFS is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GekkoFS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GekkoFS.  If not, see <https://www.gnu.org/licenses/>.

  SPDX-License-Identifier: GPL-3.0-or-later
*/

#include <filesystem>
#include <fstream>

#include <tests/config.hpp>
#include <catch2/catch.hpp>
#include <fmt/format.h>
#include <boost/process.hpp>
#include <boost/uuid/uuid.hpp>            // uuid class
#include <boost/uuid/uuid_generators.hpp> // generators
#include <boost/uuid/uuid_io.hpp>         // streaming operators etc.

#include <api/gekkofs.hpp>

using namespace std::literals;
namespace bp = boost::process;
namespace boost_fs = boost::filesystem;
namespace std_fs = std::filesystem;


// A temporary file with RAII removal
struct temporary_file {

    explicit temporary_file(std_fs::path filename)
        : filename_(std::move(filename)), ofs_(filename_) {}

    temporary_file(std_fs::path filename, const std::string_view& text)
        : filename_(std::move(filename)), ofs_(filename_) {
        write(text);
    }

    ~temporary_file() {
        ofs_.close();
        std_fs::remove(filename_);
    }

    void
    write(const std::string_view& text) {
        ofs_ << text;
        ofs_.flush();
    }

    std_fs::path
    filename() const {
        return filename_;
    }

    std_fs::path filename_;
    std::ofstream ofs_;
};

// A temporary daemon with RAII shutdown and environment removal
struct test_daemon {

    static std::tuple<std_fs::path, long>
    generate_environment() {

        static boost::uuids::random_generator generator;
        boost::uuids::uuid uuid = generator();
        boost::random::mt19937 gen;
        boost::random::uniform_int_distribution<> dist(1025, 65535);

        const auto basedir = to_string(uuid);
        const long port = dist(gen);

        CAPTURE(basedir, port);
        return {basedir, port};
    }

    test_daemon(const std_fs::path& rootdir, const std_fs::path& mountdir,
                const std::string& interface)
        : abs_rootdir(std_fs::absolute(std_fs::current_path() / rootdir)),
          abs_mountdir(std_fs::absolute(std_fs::current_path() / mountdir)) {

        auto env = boost::this_process::environment();
        env["GKFS_LOG_LEVEL"] = "100";
        env["GKFS_DAEMON_LOG_PATH"] = "/dev/stderr";

        if(!std_fs::exists(abs_rootdir)) {
            std_fs::create_directories(abs_rootdir);
        }

        if(!std_fs::exists(abs_mountdir)) {
            std_fs::create_directories(abs_mountdir);
        }

        auto search_paths = boost::this_process::path();
        search_paths.emplace_back(boost_fs::path{cmake::binary_dir} /
                                  boost_fs::path{"src/daemon"});

        daemon_ = bp::child(bp::search_path("gkfs_daemon", search_paths),
                            bp::args({"--rootdir"s, abs_rootdir, "--mountdir"s,
                                      abs_mountdir, "--listen"s, interface}));

        // since we don't have a way to check whether the daemon has finished
        // its startup process, wait for a 100ms, otherwise any connections
        // to it will fail
        std::this_thread::sleep_for(500ms);
    }

    ~test_daemon() {
        ::kill(daemon_.native_handle(), SIGTERM);

        if(!daemon_.wait_for(5s)) {
            daemon_.terminate();
        }

        std_fs::remove_all(abs_rootdir);
        std_fs::remove_all(abs_mountdir);
    }

    bp::child daemon_;
    std_fs::path abs_rootdir;
    std_fs::path abs_mountdir;
};

SCENARIO(" API connections to gkfs work as expected ", "[api][connection]") {

    // TODO: tests for connecting non-existing daemons are missing. The reason
    // TODO: is that since we don't have timeouts yet in Hermes, if the
    // TODO: daemon is missing the connection RPC hangs forever while waiting
    // TODO: for a reply that will never arrive

    using namespace std::string_literals;
    using Catch::Matchers::Contains;
    using namespace gkfs::api;

    GIVEN(" a fs::connection ") {

        WHEN(" hostfile does not exist ") {

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(),
                                    Contains("No such file or directory"s));
            }
        }

        WHEN(" hostfile is empty ") {

            temporary_file tmp{"empty.hostfile"};

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(
                        fs::connection(tmp.filename()),
                        Contains("hostfile exists but no addresses could be "
                                 "extracted"s));
            }
        }

        WHEN(" hostfile is invalid ") {

            temporary_file tmp{"invalid.hostfile",
                               "foobarbaz\nbarfoobaz\nbazbarfoo\n"};

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(tmp.filename()),
                                    Contains("invalid line: 'foobarbaz'"s));
            }
        }

        WHEN(" hostfile is partially invalid ") {

            temporary_file tmp{"pinvalid.hostfile",
                               "host.example.com "
                               "ofi+sockets://127.0.0.1:11586\n"
                               "foobarbaz\nbarfoobaz\nbazbarfoo\n"};

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(tmp.filename()),
                                    Contains("invalid line: 'foobarbaz'"s));
            }
        }

        WHEN(" hostfile is valid ") {

            const auto [testdir, port] = test_daemon::generate_environment();

            test_daemon daemon(testdir / "rootdir", testdir / "mountdir",
                               fmt::format("lo:{}", port));

            temporary_file tmp{
                    "valid.hostfile",
                    fmt::format("localhost ofi+sockets://127.0.0.1:{}\n",
                                port)};

            THEN(" a connection instance is constructed ") {
                REQUIRE_NOTHROW(fs::connection(tmp.filename()));
            }
        }
    }
}
